; 
    include \masm32\include\masm32rt.inc
; 
.686p

comment * -----------------------------------------------------
                     Build this console app with
                  "MAKEIT.BAT" on the PROJECT menu.
        ----------------------------------------------------- *

    .data
    ;;;;;;; CONFIGURATION ;;;;;;;
    ;; gu_* stands for global user preference. These are meant to be adjusted
    ;; independently of any other code.
      ; What priority should the test driver run at?
      gu_priorityClass textequ <HIGH_PRIORITY_CLASS>

      ; Number of successive runs through the test code that must take the
      ; exact same number of CPU ticks. This setting allows us to drop
      ; spurious results like context switches, interrupts, CPU throttling,
      ; cache misses and whatever other "one-off" events occur.
      ; NOTE: this number will be automatically adjusted by the test driver
      ; on an as needed basis to ensure test results are stable and timely.
      ; This is essentially the starting point for the driver.
      gu_default_sample_batch_size equ 8

      ; The driver may raise the sample_batch_size no higher than this number.
      ; Some combinations of opcodes are super stable and give spurious results
      ; at lower settings, so the driver will attempt to adjust the testing to
      ; the code sequence under test. (Currently UNIMPLEMENTED)
      gu_max_sample_batch_size equ 512

      ; How many times do we attempt to get a stable sample reading before
      ; giving up and adjusting downwards the sample_batch_size?
      gu_sample_retry_attempts equ 100000
     
      ; Number of fully stable samples we take and toss away.
      ; We do this as an attempt to stabalize the cpu at full throttle.
      ; Play with this number, the lower you take it, the less lagtime
      ; you will have before results start getting printed out. The downside
      ; is the initial numbers may not be statistically relevant.
      ; Remember the ideal situtation is to only emit numbers that are the same
      ; from one run to the next! That is, we want a standard deviation of 0!
      ; UNUSED gu_warmup_samples equ 1 ;600

      ; Total number of samples to print out. In "production" versions
      ; this ought to be some small number like 8 or 16.
      gu_total_samples equ 25
    ;;;;;;; END CONFIG ;;;;;;;;;;
    

      align 16
      g_run_cache dd gu_max_sample_batch_size dup(0)
      align 16
      g_steps dd gu_default_sample_batch_size-1
      align 16
      g_result_timings dd gu_total_samples dup(0)

      align 16
      gu_testfunction dd testit

      g_testfunction_call_count dd 0
    .code

start:
   
; 
    invoke GetCurrentProcess
    invoke SetPriorityClass, eax, gu_priorityClass
    invoke GetCurrentThread
    invoke SetThreadPriority, eax, THREAD_PRIORITY_TIME_CRITICAL
    invoke  GetCurrentProcess
    invoke  SetProcessAffinityMask,eax,1
    invoke  Sleep, 50

    print "Spinning up the processor.",13,10
    mov ecx, 31
    l_spinup:
      call init_spinup
      dec ecx
    jns l_spinup

    print "Running 4 full runs of testit_baseline.",13,10
    print "Batch Size: "
    mov eax, [g_steps]
    add eax, 1
    print ustr$(eax)
    print " ",13,10
    mov [gu_testfunction], testit_baseline
    call run_driver
    call run_driver
    call run_driver
    call run_driver

    print " ",13,10
    print "Running 4 full runs of testit, which contains two imul.",13,10
    mov [gu_testfunction], testit
    call run_driver
    call run_driver
    call run_driver
    call run_driver
    
    inkey
    exit

; 
run_driver proc
    invoke Sleep,10
    push ebx
    push esi
    push edi

    mov [g_testfunction_call_count], 0 ; initialize to zero
    
    call main
    
    print " ",13,10
    print "Min:      "
    call calc_minimum
    push eax
    print ustr$(eax)

    print " ticks ",13,10
    print "Max:      "
    call calc_maximum
    push eax
    print ustr$(eax)

    print " ticks ",13,10
    print "Range:    "
    pop eax ; max
    pop ecx ; min
    sub eax, ecx
    print ustr$(eax)

    print " ticks ",13,10
    print "Mean:     "
    call calc_mean
    push eax
    print ustr$(eax)

    print " ticks ",13,10
    print "Variance: "
    pop eax
    call calc_variance
    print ustr$(eax)

    print " ticks^2 ",13,10

    print "Batch Size: "
    mov eax, [g_steps]
    add eax, 1
    print ustr$(eax)
    print " ",13,10
    

    print "Call Count: "
    mov eax, [g_testfunction_call_count]
    print ustr$(eax)
    print " calls",13,10
    pop edi
    pop esi
    pop ebx
    ret
run_driver endp
main proc

    
    cls
    ; take some samples, but toss the results away.
    ; we do this to spin the selected core up.
    ;mov ecx, gu_warmup_samples-1
    ;test ecx,0
    ;jns l_end_prep
    ;l_prep:

    ;align 16
    ;push ecx
    ;call do_sample_run
    ;pop ecx
    ;dec ecx
    ;jns l_prep

    ;; print a { to screen to indicate that warmup is complete and
    ;; that we have actual samples getting taken.
    print " ",13,10
    print "{"
    
    l_end_prep:
    ; now run for real, printing each result to standard out.
    mov ecx, gu_total_samples-1
    align 16
l:
    call do_sample_run
    mov [g_result_timings+ecx*4], eax
    dec ecx
    jns l

    mov ecx, gu_total_samples-1
l_final_print:
    mov eax, [g_result_timings+ecx*4]
    push eax
    call print_number
    dec ecx
    jns l_final_print

    ;; now print a } to indicate that sampling is complete
    print "}",13,10
    ret
main endp

;; take samples until the cpu takes the same amount of time to do N runs.
;; RETURN: EAX=ticks taken.
align 16
do_sample_run proc
    push ecx
    push ebp

reset_and_retry:
    mov ebp, [g_steps]
    mov ecx, gu_sample_retry_attempts
    ;invoke Sleep,1000
    align 16
retry:
    
    dec ecx
    jz stepdown
    push ecx
    call spinup

    mov edi, ebp
    align 16
l:
    call timeit
    mov [g_run_cache+edi*4], eax
    dec edi
    jns l

    mov edi, [g_testfunction_call_count]
    add edi, ebp
    mov [g_testfunction_call_count], edi

    mov edi, ebp
    mov eax, [g_run_cache+edi*4]
    pop ecx
l_check_equality:
    cmp [g_run_cache+edi*4], eax
    jne retry
    dec edi
    jns l_check_equality

    pop ebp
    pop ecx
    ret
 stepdown:
    sub [g_steps], 1
    ;print "Now requiring only " ; need a better name for this... xD
    ;mov eax, [g_steps]
    ;add eax, 1
    ;print ustr$(eax)
    ;print " entries.",13,10
    jmp reset_and_retry
do_sample_run endp

;; Does not visibly change any registers.
;; Simply spins up the CPU in preperation of "real" timing tests.
align 16
init_spinup proc
    mov eax, 10
    push ecx
    mov ecx, 777777777
l:
    ;imul eax,eax,10
    dec ecx
    jns l
    pop ecx
    ret
init_spinup endp


;; Does not visibly change any registers.
;; Simply spins up the CPU in preperation of "real" timing tests.
align 16
spinup proc
    push ecx
    mov ecx, 100000
l:
    dec ecx
    jns l
    pop ecx
    ret
spinup endp



align 16
timeit proc
; warmup
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    ;rdtsc 
    ; now go for real!
    xor eax, eax
    cpuid
    rdtsc
   
    
    push eax
    push edx
    
    
    call [gu_testfunction]



    xor eax, eax
    cpuid
    rdtsc
    push eax
    push edx
    call print_runtime

    ret

timeit endp


print_runtime proc
    pop esi
    pop edx
    pop eax
    pop ecx

    cmp ecx, edx ; if RDTSC wrapped around, the high 32 bits will not be equal.
    jne wraparound
    
    pop ecx
    sub eax, ecx
    push esi
    ret
    
    align 16
wraparound:
    mov eax, 0
    pop ecx
    push esi
    ret
print_runtime endp

;; destroys the eax, edx register, not used for return result.
align 16
print_number  proc
    pop ebp
    pop eax
    push ecx
    print ustr$(eax)
    print ", ";,13,10
    pop ecx
    push ebp
    ret
print_number endp

;;;; Compute the minimum of result timings
;; result in EAX
calc_minimum proc
    mov ecx, gu_total_samples-1
    mov eax, [g_result_timings+ecx*4]
  l:
    mov edx, [g_result_timings+ecx*4]
    cmp eax, edx
    jl skip
    mov eax, edx
  skip:
    dec ecx
    jns l
    ret
calc_minimum endp

;;;; Compute the maximum of result timings
;; result in EAX
calc_maximum proc
    mov ecx, gu_total_samples-1
    mov eax, [g_result_timings+ecx*4]
  l:
    mov edx, [g_result_timings+ecx*4]
    cmp eax, edx
    jg skip
    mov eax, edx
  skip:
    dec ecx
    jns l
    ret
calc_maximum endp

;;;; Compute the mean of result timings
;; result in EAX
calc_mean proc
    mov ecx, gu_total_samples-1
    mov eax, 0
  l:
    mov edx, [g_result_timings+ecx*4]
    add eax, edx
    dec ecx
    jns l
    mov ecx, gu_total_samples
    xor edx, edx
    div ecx
    ;mov eax, edx
    ret
calc_mean endp

;;; result is eax, remainder in edx
calc_variance proc
    push edi ; callie save
    push esi ; registers
    push ebx ; all need pushed
    mov edi, eax ; store `mean` in edi for use computing variance.
    xor edx, edx
    xor eax, eax
    mov esi, gu_total_samples-1 ; tell it where we are starting...

    align 16
l:
    mov ecx, [esi*4+g_result_timings]
    sub ecx, edi      ; value - mean of all samples
    imul ecx, ecx     ; square the difference
    mov ebx, ecx
    sar ebx, 31       ; shift arithmatic right by 31
    add eax, ecx
    adc edx, ebx
    sub esi, 1
    jns l

    ;; At this point eax=low 32bits and edx=high 32 bits
    ;; esi, ebx, ecx, ebp, edi are all free for use.
    mov ecx, gu_total_samples ; number of samples to divide by
    ;;;; TODO, fix this to do proper 64 bit division!
    xor edx, edx ; we are only doing 32 bit division... discarding all high bits!
    idiv ecx

    ; now result is in eax, remainder is in edx (we usually don't care about this)
    
    pop ebx  ; restore
    pop esi  ; callie save
    pop edi  ; registers
    ret
calc_variance endp

align 16

testit_baseline proc
    mov ecx, 15
  l:
    dec ecx
    jns l
    ret
testit_baseline endp

align 16
;;;;;;;;; code under test should go here! ;;;;;;;;;
;; eax, ecx, edx, ebx are free scratch registers
testit proc
 mov ecx, 15
 l:
 
   ;
    imul eax, ecx, 1000
    imul ebx, eax, 10000
    ;pop ecx
    dec ecx
    jns l
   
ret
testit endp




; 

end start
